"""Server implements multiplex thrift server, that is, there may
exists multiple thrift services in the thrift configurarion files.

Server features are defined by a Disparcher class that implements the
methods that are triggered by clients queries, and a thrift
configuration files that define available services and the methods.
"""
# Author: Pearu Peterson
# Created: February 2019


import os
import tempfile
import socket
import time
import multiprocessing
import sys
import pickle
from . import utils

import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import thriftpy2 as thr
    import thriftpy2.rpc

try:
    import tblib
    import tblib.pickling_support
except ImportError:
    tblib = None

if tblib is not None:
    tblib.pickling_support.install()


class Processor(thr.thrift.TProcessor):

    def __init__(self, server, service, handler):
        self.server = server
        thr.thrift.TProcessor.__init__(self, service, handler)

    def handle_exception(self, e, result):
        if thr.thrift.TProcessor.handle_exception(self, e, result):
            return True
        # map Python native exception to thrift Exception so that
        # client can remap thrift Exception to Python

        exc_type = self.server.thrift.Exception
        if tblib is not None:
            exc_kind = self.server.thrift.ExceptionKind.EXC_TBLIB
            exc = exc_type(exc_kind, pickle.dumps(sys.exc_info()))
        else:
            # tblib would be required to pickle traceback instances
            exc_kind = self.server.thrift.ExceptionKind.EXC_MESSAGE
            et, ev, tb = sys.exc_info()
            exc = exc_type(exc_kind, '%s: %s' % (et.__name__, ev))
        return thr.thrift.TProcessor.handle_exception(self, exc, result)


class MultiplexedProcessor(thr.thrift.TMultiplexedProcessor):

    def __init__(self, server):
        self.server = server
        thr.thrift.TMultiplexedProcessor.__init__(self)

    def handle_exception(self, e, result):
        if thr.thrift.TProcessor.handle_exception(self, e, result):
            return True
        # map Python native exception to thrift Exception so that
        # client can remap thrift Exception to Python

        exc_type = self.server.thrift.Exception
        if tblib is not None:
            exc_kind = self.server.thrift.ExceptionKind.EXC_TBLIB
            exc = exc_type(exc_kind, pickle.dumps(sys.exc_info()))
        else:
            # tblib would be required to pickle traceback instances
            exc_kind = self.server.thrift.ExceptionKind.EXC_MESSAGE
            et, ev, tb = sys.exc_info()
            exc = exc_type(exc_kind, '%s: %s' % (et.__name__, ev))
        return thr.thrift.TProcessor.handle_exception(self, exc, result)


class Server(object):
    """Multiplex thrift server
    """

    def __init__(self, dispatcher, thrift_file, **options):
        self.multiplexed = options.pop('multiplexed', True)
        self.thrift_content_service = options.pop(
            'thrift_content_service', 'info')
        thrift_content = options.pop('thrift_content', None)
        self.options = options
        module_name = os.path.splitext(
            os.path.abspath(thrift_file))[0]+'_thrift'
        self._dispatcher = dispatcher
        self.thrift_file = thrift_file
        if thrift_content is None:
            thrift_content = utils.resolve_includes(
                open(thrift_file).read(), [os.path.dirname(thrift_file)])
        i, fn = tempfile.mkstemp(suffix='.thrift', prefix='rpc-server-')
        f = os.fdopen(i, mode='w')
        f.write(thrift_content)
        f.close()
        self.thrift = thr.load(fn, module_name=module_name)
        os.remove(fn)

    @staticmethod
    def run(dispatcher, thrift_file, options):
        """Run server in current process.
        """
        Server(dispatcher, thrift_file, **options)._serve()

    @staticmethod
    def run_bg(dispatcher, thrift_file, options, startup_time=5):
        """Run server in background process.
        """
        ctx = multiprocessing.get_context('spawn')
        p = ctx.Process(target=Server.run,
                        args=(dispatcher, thrift_file, options))
        p.start()
        start = time.time()
        while time.time() < start + startup_time:
            try:
                socket.create_connection(
                    (options['host'], options['port']), timeout=0.1)
            except ConnectionRefusedError:
                time.sleep(0.5)
            except Exception as msg:
                print('Connection failed: `%s`, trying again in 0.5 secs..'
                      % (msg))
                time.sleep(0.5)
            else:
                break
        else:
            is_alive = p.is_alive()
            if is_alive:
                p.join(1)
                p.terminate()
            raise RuntimeError(
                'failed to start up rpc_thrift server'
                ' (was alive={}, startup_time={}s)'
                .format(is_alive, startup_time))
        return p

    def _serve(self):
        """Create and run a Thrift server.
        """
        if self.multiplexed:
            service_names = [_n for _n in dir(self.thrift)
                             if not _n.startswith('_')]
            s_proc = MultiplexedProcessor(self)
            for service_name in service_names:
                service = getattr(self.thrift, service_name)
                proc = thr.thrift.TProcessor(service, self._dispatcher(self))
                s_proc.register_processor(service_name, proc)
        else:
            service = getattr(self.thrift, self.thrift_content_service)
            s_proc = Processor(self, service, self._dispatcher(self))
        server = thr.server.TThreadedServer(
            s_proc,
            thr.transport.TServerSocket(**self.options),
            iprot_factory=thr.protocol.TBinaryProtocolFactory(),
            itrans_factory=thr.transport.TBufferedTransportFactory())
        server.serve()



"""This module implements the bi-directional conversations between
objects defined by thrift module and Python objects.

One can use class instances Buffer, NDArray as annotations in the
definitions of Dispatcher methods that are decorated with
dispatchermethod.

"""
# Author: Pearu Peterson
# Created: February 2019

import pickle
import numpy as np


def fromobject(thrift, tcls, obj):
    """Create thrift object with type `tcls` from a Python object.
    """
    if isinstance(obj, tcls):
        return obj
    return globals().get(tcls.__name__, tcls)(thrift, obj)


def toobject(thrift, tobj, cls=None):
    """Convert thrift object `tobj` to Python object (with optional type
    `cls`).
    """
    if cls is not None and isinstance(tobj, cls):
        return tobj
    tcls = globals().get(type(tobj).__name__)
    if tcls is not None:
        return tcls.toobject(thrift, tobj, cls=cls)
    if cls is not None:
        return cls(tobj)
    return tobj


class Data(object):
    """Represents any data that interpretation is defined by info and kind
    attributes.

    """
    def __new__(cls, thrift, data, info=None):
        r = thrift.Data()
        if isinstance(data, str):
            r.data = data.encode()
            r.kind = thrift.DataKind.DATA_ENCODED
            r.info = 'utf-8' if info is None else info
        elif isinstance(data, bytes):
            r.data = data
            r.kind = thrift.DataKind.DATA_RAW
            r.info = ''
        else:
            # protocol = pickle.DEFAULT_PROTOCOL if info is None else int(info)
            r.data = pickle.dumps(data)
            r.kind = thrift.DataKind.DATA_PICKLED
            r.info = str(info).encode()
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        kind = obj.kind
        if kind == thrift.DataKind.DATA_ENCODED:
            encoding = obj.info
            if isinstance(obj.data, str):
                return obj.data
            return obj.data.decode(encoding)
        if kind == thrift.DataKind.DATA_RAW:
            return obj.data
        if kind == thrift.DataKind.DATA_PICKLED:
            # protocol = int(obj.info)  # pickle.loads detects protocol itself
            return pickle.loads(obj.data)
        raise NotImplementedError(repr(kind))


class Buffer(object):
    """Represents a buffer that is mapped to/from Python bytes.
    """
    def __new__(cls, thrift, data):
        r = thrift.Buffer()
        if isinstance(data, thrift.Buffer):
            data = data.data
        if isinstance(data, str):
            data = data.encode()
        elif isinstance(data, np.ndarray):
            data = data.tobytes()
        if not isinstance(data, bytes):
            raise NotImplementedError('Buffer from {}'.format(type(data)))
        r.data = data
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        data = obj.data
        if isinstance(data, str):
            if cls is str:
                return data
            data = data.encode()
        if cls is bytes:
            return data
        if cls is np.ndarray or cls is None:
            return np.frombuffer(data, dtype=np.uint8)
        raise NotImplementedError('Buffer to {}'.format(cls))


class NDArray(object):
    """Represents N-dimensional array that is mapped to/from numpy
    ndarray.
    """

    def __new__(cls, thrift, data):
        r = thrift.NDArray()
        if isinstance(data, thrift.NDArray):
            r.data = data.data
            r.typestr = data.typestr
            r.shape = data.shape
        elif isinstance(data, np.ndarray):
            r.shape = list(data.shape)
            r.typestr = data.dtype.str
            r.data = data.tobytes()   # ensures C-contiguous array
        else:
            raise NotImplementedError('Buffer from {}'.format(type(data)))
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        shape = tuple(obj.shape)
        dtype = np.dtype(obj.typestr)
        data = obj.data
        if cls is np.ndarray or cls is None:
            if isinstance(data, str):
                data = data.encode()
            # C-contiguous array
            return np.frombuffer(data, dtype=dtype).reshape(shape)
        raise NotImplementedError('NDArray to {}'.format(cls))



# Expose a temporary prototype. It will be replaced by proper
# implementation soon.
from .remotejit import RemoteJIT  # noqa: F401
from .omniscidb import RemoteOmnisci  # noqa: F401

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
